package matrices;

public class MatrixUtil {
    private MatrixUtil() {
    }
    //usuwa podany rząd i kolumnę

    public static Matrix submatrix(Matrix matrix, int row, int col) {
        Matrix mat1 = deleteRow(matrix, row);
        return deleteCol(mat1, col);
    }

    // w tablicach podane są indeksy kolumn do usunięcia
    public static Matrix submatrix(Matrix matrix, int[] rows, int[] cols) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int ii = 0;
        for (int row : rows) {
            mat = deleteRow(matrix, row - ii);
            ii++;
        }
        int jj = 0;
        for (int col : cols) {
            mat = deleteCol(mat, col - jj);
            jj++;
        }
        return mat;
    }

    public static Matrix deleteRow(Matrix matrix, int row) {
        int cc = matrix.getColCount();
        int c = cc - 1;
        double[] ar = ArrayUtil.twoToOne(matrix.getArray());
        double[] ar1 = ArrayUtil.shortenArray(ar, row * cc, row * cc + c);
        return new Matrix(ar1, cc);
    }

    public static Matrix deleteCol(Matrix matrix, int col) {
        double[][] array = ArrayUtil.clone(matrix.getArray());
        for (int i = 0; i < array.length; i++) {
            array[i] = ArrayUtil.shortenArray(array[i], col);
        }
        return new Matrix(array);
    }

    //pobiera podmacierz składającą się z podanej liczby rzędów i kolumn
    //bezpośrednio do siebie przylegających
    public static Matrix submatrix(Matrix matrix, int startRow, int endRow,
                                   int startCol, int endCol) {
        int len1 = endRow - startRow + 1;
        int len2 = endCol - startCol + 1;
        int len = len1 * len2;
        double[] ar = new double[len];
        int k = 0;
        for (int i = startRow; i < endRow + 1; i++) {
            for (int j = startCol; j < endCol + 1; j++) {
                ar[k] = matrix.getArray()[i][j];
                k++;
            }
        }
        return new Matrix(ar, len2);
    }

    public static double minor(Matrix matrix, int row, int col) {
        double det = 0.0;
        try {
            det = det(submatrix(matrix, row, col));
        } catch (MatrixException e) {
            e.printStackTrace();
        }
        return det;
    }

    public static double algComplement(Matrix matrix, int row, int col) {
        return minor(matrix, row, col) * Math.pow(-1, row + col);
    }

    public static double det(Matrix matrix) throws MatrixException {
        double det = 0.0;
        if (matrix.isSquareMatrix()) {
            switch (matrix.getRowCount()) {
                case 1:
                    det = matrix.getArray()[0][0];
                    break;
                case 2:
                    det = matrix.getArray()[0][0] * matrix.getArray()[1][1]
                            - matrix.getArray()[0][1] * matrix.getArray()[1][0];
                    break;
                case 3:
                    double a = matrix.getArray()[0][0];
                    double b = matrix.getArray()[0][1];
                    double e = matrix.getArray()[0][2];
                    double c = matrix.getArray()[1][0];
                    double d = matrix.getArray()[1][1];
                    double f = matrix.getArray()[1][2];
                    double g = matrix.getArray()[2][0];
                    double h = matrix.getArray()[2][1];
                    double i = matrix.getArray()[2][2];
                    det = ((a * d * i + b * f * g + e * c * h) - (e * d * g + a
                            * f * h + b * c * i));
                    break;
                default:
                    int row = 0;
                    for (int j = 0; j < matrix.getColCount(); j++) {
                        try {
                            det += algComplement(matrix, row, j)
                                    * matrix.getCell(row, j);
                        } catch (MatrixException e1) {
                            e1.printStackTrace();
                        }
                    }
            }
        } else {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        return det;
    }

    //Odwraca macierz m i zwraca nowa odwróconą macierz
    public static Matrix transpose(Matrix m) {
        double[][] temp = new double[m.getColCount()][m.getRowCount()];
        for (int i = 0; i < m.getRowCount(); i++) {
            for (int j = 0; j < m.getColCount(); j++) {
                temp[j][i] = m.getArray()[i][j];
            }
        }
        return new Matrix(temp);
    }

    public static Matrix multiply3(Matrix A, Matrix B) {
        int n = A.getRowCount();
        int m = A.getColCount();
        int m1 = B.getRowCount();
        int p = B.getColCount();
        if (m == m1) {
            double[][] C = new double[n][p];
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < p; j++) {
                    for (int r = 0; r < m; r++) {
                        try {
                            C[i][j] += A.getCell(i, r) * B.getCell(r, j);
                        } catch (MatrixException e) {
                            e.printStackTrace();
                        }
                    }
                }
            }
            return new Matrix(C);
        } else {
            throw new IllegalArgumentException(
                    "Nie można pomnożyć tych macierzy");
        }
    }

    public static Matrix inverseGauss(Matrix matrix) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int rows = mat.getRowCount();
        int cols = mat.getColCount();
        if (rows != cols)
            try {
                throw new MatrixException("Macierz nie jest kwadratowa");
            } catch (MatrixException e) {
                e.printStackTrace();
            }
        Matrix temp = new Matrix(rows, cols * 2);
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                setCell(temp, i, j, getCell(mat, i, j));
                setCell(temp, i, j + cols, i == j ? 1 : 0);
            }
        }
        temp = gaussElimination(temp);
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                if (getCell(temp, i, j) != (i == j ? 1 : 0)) {
                    try {
                        throw new MatrixException("Macierz nie ma odwrotności");
                    } catch (MatrixException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++)
                setCell(mat, i, j, getCell(temp, i, j + cols));
        }
        Matrix mat1 = new Matrix(rows, cols);
        for (int m = 0; m < rows; m++) {
            for (int n = 0; n < cols; n++) {
                setCell(mat1, m, n, getCell(temp, m, n + cols));
            }
        }
        return mat1;
    }

    public static Matrix gaussElimination(Matrix matrix) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int rows = mat.getRowCount();
        int cols = mat.getColCount();
        int np = 0;
        for (int j = 0; j < cols; j++) {
            int pr = np;
            while (pr < rows && getCell(mat, pr, j) == 0) {
                pr++;
            }
            if (pr == rows) {
                continue;
            }
            operT(mat, Elem.Row, np, pr);
            pr = np;
            np++;
            operI(mat, Elem.Row, pr, 1.0 / getCell(mat, pr, j));
            for (int i = pr + 1; i < rows; i++) {
                operE(mat, Elem.Row, i, pr, -1 * getCell(mat, i, j));
            }
        }
        for (int i = rows - 1; i >= 0; i--) {
            int pc = 0;
            while (pc < cols && getCell(mat, i, pc) == 0) {
                pc++;
            }
            if (pc == cols) {
                continue;
            }
            for (int j = i - 1; j >= 0; j--) {
                operE(mat, Elem.Row, j, i, -1 * getCell(mat, j, pc));
            }
        }
        return mat;
    }

    public static double getCell(Matrix matrix, int row, int col) {
        return matrix.getArray()[row][col];
    }

    public static void setCell(Matrix matrix, int row, int col, double value) {
        try {
            matrix.setCell(row, col, value);
        } catch (MatrixException e) {
            e.printStackTrace();
        }
    }

    //zamienia wiersze lub kolumny o wskazanych indeksach
    //zwraca liczbę zamian czyli 1
    public static int operT(Matrix matrix, Elem elem, int index1, int index2) {
        switch (elem) {
            case Col:
                matrix = swapCols(matrix, index1, index2);
                break;
            case Row:
                matrix = swapRows(matrix, index1, index2);
                break;
        }
        return 1;
    }

    //indeks1 to indeks wiersza/kolumny, do którego dodajemy
    //indeks2 to indeks wiersza/kolumny mnożonego i dodawanego
    //c - mnożnik
    //elem - określa czy chodzi o wiersze czy o kolumny
    public static void operE(Matrix matrix, Elem elem, int index1, int index2,
                             double c) {
        switch (elem) {
            case Col:
                double[] col = matrix.getCol(index2);
                for (int i = 0; i < col.length; i++) {
                    col[i] *= c;
                }
                double[] col1 = matrix.getCol(index1);
                for (int j = 0; j < col1.length; j++) {
                    col1[j] = col1[j] + col[j];
                }
                try {
                    matrix.setCol(index1, col1);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                break;
            case Row:
                double[] row = matrix.getRow(index2);
                for (int k = 0; k < row.length; k++) {
                    row[k] *= c;
                }
                double[] row1 = matrix.getRow(index1);
                for (int m = 0; m < row1.length; m++) {
                    row1[m] = row1[m] + row[m];
                }
                try {
                    matrix.setRow(index1, row1);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                break;
        }
    }

    //wykonuje operacje mnożenia na podanym wierszu lub kolumnie
    //zwraca mnożnik, czyli c
    public static double operI(Matrix matrix, Elem elem, int index, double c) {
        switch (elem) {
            case Col:
                double[] col = matrix.getCol(index);
                for (int i = 0; i < col.length; i++) {
                    if (col[i] != 0) {
                        col[i] *= c;
                    }
                }
                try {
                    matrix.setCol(index, col);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                break;
            case Row:
                double[] row = matrix.getRow(index);
                for (int i = 0; i < row.length; i++) {
                    if (row[i] != 0) {
                        row[i] *= c;
                    }
                }
                try {
                    matrix.setRow(index, row);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                break;
        }
        return c;
    }

    public static Matrix swapCols(Matrix matrix, int col1, int col2) {
        double[] c1 = matrix.getCol(col1).clone();
        double[] c2 = matrix.getCol(col2).clone();
        try {
            matrix.setCol(col2, c1);
            matrix.setCol(col1, c2);
        } catch (MatrixException e) {
            e.printStackTrace();
        }
        return matrix;
    }

    public static Matrix swapRows(Matrix matrix, int row1, int row2) {
        double[] r1 = matrix.getRow(row1).clone();
        double[] r2 = matrix.getRow(row2).clone();
        try {
            matrix.setRow(row2, r1);
            matrix.setRow(row1, r2);
        } catch (MatrixException e) {
            e.printStackTrace();
        }
        return matrix;
    }

    public static Matrix pow(Matrix matrix, int times) {
        if (matrix.isSquareMatrix()) {
            Matrix t = null;
            try {
                t = matrix.clone();
            } catch (CloneNotSupportedException e1) {
                e1.printStackTrace();
            }
            for (int i = 0; i < times - 1; i++) {
                try {
                    t = t.multiply2(matrix.clone());
                } catch (CloneNotSupportedException e) {
                    e.printStackTrace();
                }
            }
            return t;
        }
        return null;
    }

    public static double detGauss(Matrix matrix) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int sw = 0;
        int rows = mat.getRowCount();
        int cols = mat.getColCount();
        for (int c = 0; c < cols; c++) {
            for (int r = rows - 1; r > c; r--) {
                int a = countStartZeroes(mat, Elem.Row, r - 1);
                int b = countStartZeroes(mat, Elem.Row, r);
                if (a > b) {
                    sw += operT(mat, Elem.Row, r - 1, r);
                } else if (a == b) {
                    double w1 = getCell(mat, r - 1, c);
                    double w2 = getCell(mat, r, c);
                    operE(mat, Elem.Row, r, r - 1, -w2 / w1);
                }
            }
        }
        double[] temp = null;
        temp = mat.getDiagonal();
        double det = 1.0;
        for (double v : temp) {
            det *= v;
        }
        if ((sw & 1) == 1) {
            det *= -1.0;
        }
        return det;
    }

    public static int countStartZeroes(Matrix matrix, Elem elem, int index) {
        int sum = 0;
        double[] tabl = null;
        switch (elem) {
            case Col:
                tabl = matrix.getCol(index);
                break;
            case Row:
                tabl = matrix.getRow(index);
                break;
        }
        for (double v : tabl) {
            if (v == 0) {
                sum++;
            } else {
                return sum;
            }
        }
        return sum;
    }

    //Redukuje macierz do macierzy schodkowej lub górnotrójkątnej
    //używając operacji elementarnych
    //operE nie zmienia wyznacznika
    //operT zmienia znak wyznacznika
    //operI mnoży wyznacznik
    public static Matrix elimGauss(Matrix matrix) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int rows = mat.getRowCount();
        int cols = mat.getColCount();
        for (int c = 0; c < cols; c++) {
            for (int r = rows - 1; r > c; r--) {
                int a = countStartZeroes(mat, Elem.Row, r - 1);
                int b = countStartZeroes(mat, Elem.Row, r);
                if (a > b) {
                    operT(mat, Elem.Row, r - 1, r);
                } else if (a == b) {
                    double w1 = getCell(mat, r - 1, c);
                    double w2 = getCell(mat, r, c);
                    operE(mat, Elem.Row, r, r - 1, -w2 / w1);
                }
            }
        }
        return mat;
    }

    public static double perm(Matrix matrix) throws MatrixException {
        double det = 0.0;
        if (matrix.isSquareMatrix()) {
            switch (matrix.getRowCount()) {
                case 1:
                    det = matrix.getArray()[0][0];
                    break;
                case 2:
                    det = matrix.getArray()[0][0] * matrix.getArray()[1][1]
                            + matrix.getArray()[0][1] * matrix.getArray()[1][0];
                    break;
                case 3:
                    double a = matrix.getArray()[0][0];
                    double b = matrix.getArray()[0][1];
                    double e = matrix.getArray()[0][2];
                    double c = matrix.getArray()[1][0];
                    double d = matrix.getArray()[1][1];
                    double f = matrix.getArray()[1][2];
                    double g = matrix.getArray()[2][0];
                    double h = matrix.getArray()[2][1];
                    double i = matrix.getArray()[2][2];
                    det = ((a * d * i + b * f * g + e * c * h) + (e * d * g + a
                            * f * h + b * c * i));
                    break;
                default:
                    int row = 1;
                    for (int j = 0; j < matrix.getColCount(); j++) {
                        try {
                            det += algComplement2(matrix, row, j)
                                    * matrix.getCell(row, j);
                        } catch (MatrixException e1) {
                            e1.printStackTrace();
                        }
                    }
                    break;
            }
        } else {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        return det;
    }

    public static double algComplement2(Matrix matrix, int row, int col) {
        return minor(matrix, row, col) * Math.pow(1, row + col);
    }

    public static Matrix inverse(Matrix matrix) {
        int r = matrix.getRowCount();
        int c = matrix.getColCount();
        Matrix matrixac = new Matrix(r, c);
        for (int i = 0; i < r; i++) {
            for (int j = 0; j < c; j++) {
                double ac = algComplement(matrix, i, j);
                try {
                    matrixac.setCell(i, j, ac);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
            }
        }
        Matrix matrix1 = transpose(matrixac);
        double det = 0.0;
        try {
            det = matrix.det();
        } catch (MatrixException e) {
            e.printStackTrace();
        }
        return matrix1.mult2(1.0 / det);
    }

    //------------------ eliminacja Gaussa ----------------------------------
    //indeks1 to indeks wiersza/kolumny, do którego dodajemy
    //indeks2 to indeks wiersza/kolumny mnożonego i dodawanego
    //c - mnożnik
    //elem - określa czy chodzi o wiersze czy o kolumny
    public static Matrix makeE(Matrix matrix, Elem elem, int index1,
                               int index2, double c) {
        int i = matrix.getRowCount();
        int j = matrix.getColCount();
        double[][] arr = new double[i][j];
        switch (elem) {
            case Row:
                for (int r = 0; r < i; r++) {
                    for (int k = 0; k < j; k++) {
                        if ((r == index1) && (k == index2)) {
                            arr[r][k] = c;
                        } else if (r == k) {
                            arr[r][k] = 1;
                        } else {
                            arr[r][k] = 0;
                        }
                    }
                }
                break;
            case Col:
                for (int r = 0; r < i; r++) {
                    for (int k = 0; k < j; k++) {
                        if ((r == index2) && (k == index1)) {
                            arr[r][k] = c;
                        } else if (r == k) {
                            arr[r][k] = 1;
                        } else {
                            arr[r][k] = 0;
                        }
                    }
                }
                break;
        }
        return new Matrix(arr);
    }

    //zamienia wiersz o index1 na wiersz index2 i odwrotnie
    public static Matrix makeT(Matrix matrix, int index1, int index2) {
        int i = matrix.getRowCount();
        int j = matrix.getColCount();
        double[][] arr = new double[i][j];
        for (int r = 0; r < i; r++) {
            for (int k = 0; k < j; k++) {
                if ((r == k) && (r != index1) && (r != index2)
                        || ((r == index1) && (k == index2))
                        || ((r == index2) && (k == index1))) {
                    arr[r][k] = 1;
                } else {
                    arr[r][k] = 0;
                }
            }
        }
        return new Matrix(arr);
    }

    //index to indeks wiersza mnożonego przez c
    //c - mnożnik
    public static Matrix makeI(Matrix matrix, int index, double c) {
        int i = matrix.getRowCount();
        int j = matrix.getColCount();
        double[][] arr = new double[i][j];
        for (int r = 0; r < i; r++) {
            for (int k = 0; k < j; k++) {
                if ((r == k) && (r == index)) {
                    arr[r][k] = c;
                } else if ((r == k) && (r != index)) {
                    arr[r][k] = 1;
                } else {
                    arr[r][k] = 0;
                }
            }
        }
        return new Matrix(arr);
    }

    public static Matrix naiveGauss(Matrix matrix) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int n = mat.getRowCount();
        for (int k = 0; k <= n - 1; k++) {
            for (int i = k + 1; i < n; i++) {
                double x = getCell(mat, i, k) / getCell(mat, k, k);
                try {
                    mat.setCell(i, k, 0);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                for (int j = k + 1; j < n; j++) {
                    double val = getCell(mat, i, j) - x * getCell(mat, k, j);
                    try {
                        mat.setCell(i, j, val);
                    } catch (MatrixException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        return mat;
    }

    public static Matrix permutGauss(Matrix matrix, int[] perm) {
        Matrix mat = null;
        try {
            mat = matrix.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        int n = mat.getRowCount();
        for (int k = 0; k <= n - 1; k++) {
            for (int i = k + 1; i < n; i++) {
                double z = getCell(mat, perm[i], k) / getCell(mat, perm[k], k);
                setCell(mat, perm[i], k, 0);
                for (int j = k + 1; j < n; j++) {
                    setCell(mat, perm[i], j, getCell(mat, perm[i], j) - z
                            * getCell(mat, perm[k], j));
                }
            }
        }
        return mat;
    }
}
